'''
Defines the base class for all choosers in LSR.

@author: Peter Parente
@organization: IBM Corporation
@copyright: Copyright (c) 2005, 2006 IBM Corporation
@license: Common Public License 1.0

All rights reserved. This program and the accompanying materials are made 
available under the terms of the Common Public License v1.0 which accompanies
this distribution, and is available at
U{http://www.opensource.org/licenses/cpl1.0.php}
'''
import weakref, os.path
import AEEvent, UIElement

SINGLETONS = weakref.WeakValueDictionary()

class AEChooser(UIElement.UIE):
  '''  
  Most abstract base class for all L{AEChooser} dialogs.
  
  This class is abstract as most of its methods raise NotImplementedError and 
  need to be overriden in subclasses.
  
  @cvar SINGLETON: Should only one instance of this chooser live at a time? 
    This flag should be overriden by a subclass to indicate whether the
    chooser is a singleton or not.
  @type SINGLETON: boolean
  @cvar OK: Indicates the chooser is completing and its options should be
    applied
  @type OK: integer
  @cvar CANCEL: Indicates the chooser is canceling and its options should be
    ignored
  @type CANCEL: integer
  @cvar APPLY: Indicates the chooser options should be applied immediately 
    with no changes to its state
  @type APPLY: integer
  @ivar event_man: Reference to the event manager to which an event will be
    posted when the user indicates the dialog is completed, cancelled, or
    the current optionss applied
  @type event_man: L{EventManager}
  @ivar aid: Unique identifier for the application L{Tier} with which the 
    L{AEChooser} that fired this event is associated
  @type aid: opaque
  @ivar block_signals: Blocks future signals from being sent by the L{_signal}
    method after a L{OK} or L{CANCEL} signal is sent
  @type block_signals: boolean
  '''
  SINGLETON = False
  # these must be mutually exclusive from the signal constants used by subclass
  CANCEL = -1000
  APPLY = -1001
  OK = -1002
  
  def __init__(self):
    '''
    Initializes the L{event_man} and L{aid} variables to None and sets the 
    L{block_signals} flag to False.
    '''
    self.aid = None
    self.event_man = None
    self.block_signals = False
    
  def __call__(self, event_man, aid):
    '''    
    Stores a reference the L{EventManager} that will be notified by L{_signal}
    and a reference to the L{Tier} in which the L{Perk} associated with this
    L{AEChooser} is loaded.
    
    @param event_man: Reference to the event manager to which an event will be
      posted when the user indicates the dialog is completed, cancelled, or
      the current options applied
    @type event_man: L{EventManager}
    @param aid: Unique identifier for the application L{Tier} with which the 
      L{AEChooser} that fired this event is associated
    @type aid: L{Tier}
    @raise ValueError: When the L{SINGLETON} flag is set to True and a 
      singleton already exists
    '''
    cls = self.getClassName()
    if self.SINGLETON and cls in SINGLETONS:
      raise ValueError(SINGLETONS[cls])
    self.aid = aid
    self.event_man = event_man
    # store a weakref for singletons
    if self.SINGLETON:
      SINGLETONS[cls] = self
  
  def init(self, **kwargs):
    '''
    Initializes the chooser. Should enable the chooser for interaction with the
    user.
    
    @raise NotImplementedError: When not overridden by a subclass
    '''
    raise NotImplementedError
  
  def activate(self, **kwargs):
    '''
    Activates the chooser. Called when a singleton instance of the chooser
    already exists, but an attempt was just made to create another. Does 
    nothing by default. Could be used to, for instance, raise a window to the
    foreground.
    '''
    pass
    
  def close(self):
    '''
    Closes the chooser. Should prevent further chooser interaction with the 
    user.
    
    @raise NotImplementedError: When not overridden by a subclass
    '''
    raise NotImplementedError
  
  def update(self, **kwargs):
    '''
    Does an update of some aspect of the L{AEChooser}. A subclass can override
    this method to support updates from observers of L{AEEvent.ChooserChange}
    events.
    
    @param kwargs: Arbitrary data given by the observer. The L{AEChooser}
      implementor should strong-name keyword params of interest.
    @type kwargs: dictionary
    '''
    pass
  
  def _getResource(self, name):
    '''
    Gets the absolute path to a file located in the same directory as this
    L{AEChooser}. Useful for locating resources like glade files.
    
    @return: Path to a resource file in the same directory as the chooser
    @rtype: string
    '''
    return os.path.join(self.getPath(), name)
  
  def _signal(self, kind, **kwargs):
    '''    
    Posts an L{AEEvent.ChooserChange} event to the L{EventManager}. The kind of
    change should be any arbitrary string or one of L{OK}, L{APPLY}, L{CANCEL}.
    Any keyword arguments will be delivered to an observer. The event will also
    include a reference to this L{AEChooser} such that an observer can call the
    L{update} method on it. After one cancel or one OK signal is propogated,
    no more signals are allowed to be sent.
    
    @param kind: Kind of signal, one of L{OK}, L{APPLY}, L{CANCEL} or a chooser
      defined integer constant
    @type kind: integer
    @param kwargs: Arbitrary data to be delivered with the event
    @type kwargs: dictionary
    '''
    if self.block_signals:
      return
    self.event_man.postEvents(AEEvent.ChooserChange(self.aid, self, kind, 
                                                    **kwargs))
    if kind == self.CANCEL or kind == self.OK:
      self.block_signals = True
