'''
Defines a L{Task} to execute in response to a L{AEEvent.ViewChange}.

@author: Peter Parente
@organization: IBM Corporation
@copyright: Copyright (c) 2005, 2006 IBM Corporation
@license: Common Public License 1.0

All rights reserved. This program and the accompanying materials are made 
available under the terms of the Common Public License v1.0 which accompanies
this distribution, and is available at
U{http://www.opensource.org/licenses/cpl1.0.php}
'''

import EventTask, Constants
import AEEvent

class ViewTask(EventTask.EventTask):
  '''
  Executed when a view change occurs.
  
  This class registers its name and whether it should be monitored by default
  in an L{AEMonitor} using the L{Constants.registerTaskType} function when this
  module is first imported. An L{AEMonitor} can use this information to build
  its menus.
  
  @cvar GAINED: View was gained by activating it
  @type GAINED: integer
  @cvar LOST: View was lost by deactivating it
  @type LOST: integer
  @cvar STARTUP: View was identified at LSR startup
  @type STARTUP: integer
  @cvar FIRST_GAINED: View was gained by activating it during LSR startup
  @type FIRST_GAINED: integer
  '''
  LOST = AEEvent.ViewChange.LOST
  GAINED = AEEvent.ViewChange.GAINED
  STARTUP = AEEvent.ViewChange.STARTUP
  FIRST_GAINED = AEEvent.ViewChange.FIRST_GAINED
  Constants.registerTaskType('ViewTask', True)
  
  def getEventType(self):
    '''
    @return: Type of L{AEEvent} this L{Task} wants to handle
    @rtype: class
    '''
    return AEEvent.ViewChange
  
  def update(self, por, title, gained, layer, **kwargs):
    '''
    Updates this L{Task} in response to a consumed view change event. Called by 
    L{Tier.Tier._executeTask}.
    
    @param por: Point of regard to the root of the new view
    @type por: L{POR}
    @param title: Title of the view
    @type title: string
    @param gained: One of L{LOST}, L{GAINED}, L{STARTUP}, L{FIRST_GAINED}
    @type gained: integer
    @param layer: Layer on which the event occurred, one of
      L{Task.LAYER_FOCUS}, L{Task.LAYER_TIER}, or L{Task.LAYER_BACKGROUND}
    @type layer: integer
    '''
    pass

  def execute(self, por, title, gained, layer, **kwargs):
    '''
    Executes this L{Task} in response to a view change event. Called by 
    L{Tier.Tier._executeTask}.
    
    @param por: Point of regard to the root of the new view
    @type por: L{POR}
    @param title: Title of the view
    @type title: string
    @param gained: One of L{LOST}, L{GAINED}, L{STARTUP}, L{FIRST_GAINED}
    @type gained: integer
    @param layer: Layer on which the event occurred, one of
      L{Task.LAYER_FOCUS}, L{Task.LAYER_TIER}, or L{Task.LAYER_BACKGROUND}
    @type layer: integer
    @return: True to allow other L{Task}s to process this event
    @rtype: boolean
    '''
    if gained == self.GAINED:
      return self.executeGained(por=por, title=title, layer=layer, **kwargs)
    elif gained == self.LOST:
      return self.executeLost(por=por, title=title, layer=layer, **kwargs)
    elif gained == self.STARTUP:
      return self.executeStartup(por=por, title=title, layer=layer, **kwargs)
    elif gained == self.FIRST_GAINED:
      return self.executeFirstGained(por=por, title=title, layer=layer, 
                                     **kwargs)
    
  def executeGained(self, por, title, layer, **kwargs):
    '''
    Executes this L{Task} in response to a view gained event. Called by 
    L{execute}.
    
    @param por: Point of regard to the root of the new view
    @type por: L{POR}
    @param title: Title of the view
    @type title: string
    @param layer: Layer on which the event occurred, one of
      L{Task.LAYER_FOCUS}, L{Task.LAYER_TIER}, or L{Task.LAYER_BACKGROUND}
    @type layer: integer
    @return: True to allow other L{Task}s to process this event
    @rtype: boolean
    '''
    return True
  
  def executeLost(self, por, title, layer, **kwargs):
    '''    
    Executes this L{Task} in response to a view lost event. Called by
    L{execute}.
    
    @param por: Point of regard to the root of the lost view
    @type por: L{POR}
    @param title: Title of the view
    @type title: string
    @param layer: Layer on which the event occurred, one of
      L{Task.LAYER_FOCUS}, L{Task.LAYER_TIER}, or L{Task.LAYER_BACKGROUND}
    @type layer: integer
    @return: True to allow other L{Task}s to process this event
    @rtype: boolean
    '''
    return True
  
  def executeStartup(self, por, title, layer, **kwargs):
    '''
    Executes this L{Task} in response to the view being identified during LSR
    startup. Called by L{execute}.
    
    @param por: Point of regard to the root of the mutated view
    @type por: L{POR}
    @param title: Title of the view
    @type title: string
    @param layer: Layer on which the event occurred, one of
      L{Task.LAYER_FOCUS}, L{Task.LAYER_TIER}, or L{Task.LAYER_BACKGROUND}
    @type layer: integer
    @return: True to allow other L{Task}s to process this event
    @rtype: boolean
    '''
    return True

  def executeFirstGained(self, por, title, layer, **kwargs):
    '''
    Executes this L{Task} in response to the view being identified as the 
    active view during LSR startup. Called by L{execute}.
    
    @param por: Point of regard to the root of the mutated view
    @type por: L{POR}
    @param title: Title of the view
    @type title: string
    @param layer: Layer on which the event occurred, one of
      L{Task.LAYER_FOCUS}, L{Task.LAYER_TIER}, or L{Task.LAYER_BACKGROUND}
    @type layer: integer
    @return: True to allow other L{Task}s to process this event
    @rtype: boolean
    '''
    return True